// DropTarget.cpp: implementation of the CDropTarget class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "dropDlg.h"
#include "DropTarget.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// IDropTarget implementation

// This implementation is tightly linked with the sample dialog. A small child window with a
// target on it is the drop "target", but this in turn relegates the drop to whatever shell
// object is specified in the "Target path" box in the dialog. When our DragEnter() is called,
// we try to obtain the IDropTarget of the shell item, and from that point on we redirect all
// our methods to it. So when the OLE subsystem inquires whether we can accept the data, it
// is the target shell item that decides what should be done, NOT our minimal implementation.

// helper function to parse pathnames for PIDLs
LPITEMIDLIST GetPIDLFromPath(LPCTSTR pszPath, LPSHELLFOLDER psfRelative)
{
   // IShellFolder::ParseDisplayName requires the path name in Unicode.
	LPWSTR pWidePath;
#ifdef UNICODE
   pWidePath = pszPath; // argument already wide
#else
	OLECHAR olePath[MAX_PATH]; // wide-char version of path name
   MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, pszPath, -1, olePath, MAX_PATH);
	pWidePath = olePath;
#endif

   LPITEMIDLIST pidl = NULL; // relative to psfRelative
   HRESULT hr = psfRelative->ParseDisplayName(NULL, NULL, pWidePath, NULL, &pidl, NULL);

	return pidl; // callers should free this
}

// replacement helper for the missing SHBindToParent() API, jeez!
HRESULT MySHBindToParent(LPCITEMIDLIST fullPidl, REFIID riid, VOID **ppv, LPCITEMIDLIST *ppidlLast)
{
	HRESULT hr;
   LPSHELLFOLDER psfDesktop = NULL; // get desktop folder again, what a waste!
	hr = SHGetDesktopFolder(&psfDesktop);
	ASSERT(psfDesktop);

	// chop up the pidl excluding the last portion, to get the parent's pidl
	LPITEMIDLIST pidl = (LPITEMIDLIST)fullPidl, pidlLast;
	ASSERT(pidl->mkid.cb); // shouldn't be an empty pidl this
	do {
		pidlLast = pidl;
		// jump to the next SHITEMID token
		pidl = (LPITEMIDLIST) ((char*)pidl + pidl->mkid.cb);
	} while(pidl->mkid.cb); // till we hit the terminator

	USHORT cbSave = pidlLast->mkid.cb;
	pidlLast->mkid.cb = 0;
	// now fullPidl is one token short, pointing to the parent folder

	// ask for the specified interface
	*ppv = NULL;
	hr = psfDesktop->BindToObject(fullPidl, NULL, riid, ppv);

	// restore the original pidl
	pidlLast->mkid.cb = cbSave;
	// pidlLast points to the relative pidl of the item
	// pass it back to the caller, if requested
	if(ppidlLast) *ppidlLast = pidlLast;
	// NOTE: this is NOT a newly allocated pidl so it shouldn't be destroyed!

	psfDesktop->Release();
	// callers should release ppv, if valid
	return hr;
}

// debug helper, emiting the format names in a data object
void DumpDataObject(LPDATAOBJECT pDO)
{
	IEnumFORMATETC* pEnum;
	HRESULT hr = pDO->EnumFormatEtc(DATADIR_GET, &pEnum);
	if(pEnum) {
		TRACE(_T("Dumping data object 0x%p...\n"), pDO);

		FORMATETC fets;
		TCHAR buf[MAX_PATH];
		while(S_OK == pEnum->Next(1, &fets, NULL)) {
			// CF_HDROP=15 it won't return a name...
			GetClipboardFormatName(fets.cfFormat, buf, sizeof(buf));
			CString str;
			str.Format(_T("Format name: \"%s\" (%d)\n"), buf, fets.cfFormat);
			TRACE(_T("%s"), (LPCTSTR)str);
		}

		pEnum->Release();
	}
	
}

// first called when the mouse enters our target window space
STDMETHODIMP CDropTarget::DragEnter(LPDATAOBJECT pDataObject, DWORD grfKeyState, POINTL pt, LPDWORD pdwEffect)
{
#ifdef _DEBUG
	TRACE(_T("In DragEnter(), examining source data object\n"));
	DumpDataObject(pDataObject); // see what's in store for us

	// estimate the current reference counter on the dragged object
	// i'm keeping hold of the object, making sure that it won't be released too early
	// this should be released somewhere before the end
	m_xrefDataObj_ = pDataObject->AddRef() - 1/*don't count our extra reference*/; 
	m_pDataObj_ = pDataObject; // keep a pointer to it for DragLeave()
#endif

	BOOL ok = FALSE;
	// don't set "*pdwEffect = DROPEFFECT_NONE"; that would adversely affect the drop target

	// attempt to obtain the IDropTarget of our nominated target
	ASSERT(m_pIDTFolder==NULL); // no leftovers from previous drops, surely!
	ASSERT(m_pwDialog);
	CString strPath;
	m_pwDialog->GetDlgItem(IDC_FOLDER)->GetWindowText(strPath);
	// if the path is empty, then we can't accept anything
	if(strPath.IsEmpty()) return S_OK;

	// try to obtain the fully qualified PIDL (from the desktop)
   HRESULT hr;
   LPMALLOC pMalloc = NULL; // memory manager, for freeing up PIDLs
   hr = SHGetMalloc(&pMalloc);
	LPSHELLFOLDER psfDesktop = NULL; // namespace root for parsing the path
	hr = SHGetDesktopFolder(&psfDesktop);
	ASSERT(pMalloc && psfDesktop); // unlikely to fail, these two

	LPITEMIDLIST fullPidl = GetPIDLFromPath(strPath, psfDesktop);
	if(fullPidl) {
		// we need to GetUIObjectOf on ourselves, so we need the parent shell folder
		LPSHELLFOLDER psfParent = NULL;
		LPCITEMIDLIST pidlRelative = NULL; // the item's PIDL relative to the parent folder
		
		// not only do we get the parent folder, we also get our relative PIDL, cool!
		// I just realized that SHBindToParent() isn't supported even for win98!
		hr = MySHBindToParent(fullPidl, IID_IShellFolder, (void**)&psfParent, &pidlRelative);
		ASSERT(psfParent && pidlRelative);

		// finally, ask for the IDropTarget of this item
		// plain files won't have any, but folders and ZIP files most certainly do!
		hr = psfParent->GetUIObjectOf(NULL, 1, (LPCITEMIDLIST*)&pidlRelative, 
			IID_IDropTarget, NULL, (void**)&m_pIDTFolder);
		if(m_pIDTFolder) {
			// this interface pointer will be held till this drag loop is terminated

			// relay the call to the shell item
			hr = m_pIDTFolder->DragEnter(pDataObject, grfKeyState, pt, pdwEffect);
			if(SUCCEEDED(hr)) ok = TRUE;
			// if the target can understand the data object, it will set pdwEffect apropriately
		}

		psfParent->Release();
		//pMalloc->Free(pidlRelative); // not to be freed! we get its offset within "fullPidl"
		pMalloc->Free(fullPidl);
	}
	// else most probably we have an invalid path, drop is impossible

	// cleanup intermediate interfaces
	psfDesktop->Release();
	pMalloc->Release();

	if(!ok) {
		*pdwEffect = DROPEFFECT_NONE; // we can't understand this thing
		TRACE(_T("DragEnter failed for this or the other reason\n"));
	}
	return S_OK;
}

// from now on things get really simple
STDMETHODIMP CDropTarget::DragOver(DWORD grfKeyState, POINTL pt, LPDWORD pdwEffect)
{
	// if we have a valid target object, relay the call
	if(m_pIDTFolder)
		// mouse pointer information is not really relevant to the shell item...
		return m_pIDTFolder->DragOver(grfKeyState, pt, pdwEffect);
	else {
		*pdwEffect = DROPEFFECT_NONE; // can't accept nowt
		return S_OK;
	}
}

// Either DragLeave or Drop will be called for finishing the operation, but NOT both
// DragLeave is for when the mouse leaves the target, or the drag is somehow cancelled
STDMETHODIMP CDropTarget::DragLeave()
{
	TRACE(_T("In DragLeave()\n"));

	HRESULT hr;
	if(m_pIDTFolder) {
		// next call should release the shell item's grip on the transferred data object
		hr = m_pIDTFolder->DragLeave();
		// and we are done with the drop target interface, too
		m_pIDTFolder->Release();
		m_pIDTFolder = NULL;
	}
	else hr = S_OK;

#ifdef _DEBUG
	// the reference counter on the data object should now be equal to the original
	DWORD cnt = m_pDataObj_->Release();
	// as it turns out, the dataObject is fully released by OLE when we get in here!!!
	ASSERT(m_xrefDataObj_ == cnt+1);
#endif

	return hr;
}

// exit point when something is actually dropped
STDMETHODIMP CDropTarget::Drop(LPDATAOBJECT pDataObject, DWORD grfKeyState, POINTL pt, LPDWORD pdwEffect)
{
	TRACE(_T("In Drop(), effect=%ld\n"), *pdwEffect);

	ASSERT(m_pIDTFolder); // we can't be called if we kept on returning DROPEFFECT_NONE
	ASSERT(*pdwEffect != DROPEFFECT_NONE); // that goes without saying

	HRESULT hr;
	// this call will actually perform the file operation in pdwEffect
	hr = m_pIDTFolder->Drop(pDataObject, grfKeyState, pt, pdwEffect);

	// all is said & done
	// perform cleanup similar to DragLeave() above
	m_pIDTFolder->Release();
	m_pIDTFolder = NULL;

#ifdef _DEBUG
	// the above drop could have added a format or two to the original object, tsekit
	DumpDataObject(pDataObject);

	// the reference counter on the data object should now be equal to the original
	ASSERT(pDataObject == m_pDataObj_);
	DWORD cnt = m_pDataObj_->Release();
	// contrary to DragLeave() here OLE still holds onto the data object
	ASSERT(m_xrefDataObj_ == cnt); // perfect agreement
#endif

	return hr;
}
